#!/usr/bin/perl
#
#  Copyright 2000, G. Wade Johnson
#  Released under the same license as Perl.
#  
#  Version: 0.7.2
#
use Getopt::Std;
use LWP::Simple;
use HTML::Parser;
use Data::Dumper;
use URI::URL;
use XML::Data::DataDoc;
use strict;

{
 package  HTML2DataDoc;
 @HTML2DataDoc::ISA = ( 'HTML::Parser' );

# Added attributes to HTML::Parser
#   curr        - current element
#   text        - text accumulator
#   get_content - flag for whether or not we want content
#   siteinfo    - the siteinfo element for the current website
#   website     - the current website
#   resources   - the current resources list
#   navigation  - the current navigation list
#   page        - the current page
#   group       - the current group
#   set         - the current set
#   url         - the current url
#   cntr        - hash of counters for different ids
#   pagelist    - hash of pages we've visited by url
#   stylelist   - hash of stylesheets we've seen by url
#   contactlist - hash of contacts we've seen by email address
#   grouplist   - hash of groups we've created by root
#   imglist     - hash of images we've seen by url
#   queue       - list of pages to visit

 ##--------------------------------
 #  parse the url given as a first argument into the current DataDoc.
 sub  parse_url
  {
   my  $self = shift;
   my  $url  = shift;

   my $base = $url;
   $base =~ s#[^/]+$##;

   $self->_create_website( $base ) unless($self->{website});

   $self->{url}  = $url;
   $self->{base} = $base;

   my $page = LWP::Simple::get( $url );
   if($page)
    {
     $self->parse( $page );
    }
   else
    {
     warn "Error retrieving <$url>.\n"
    }
  }


 ##--------------------------------
 #  create a website element in the current parser.
 sub _create_website
  {
   my $self = shift;
   my $base = shift;

   my $ws = new XML::Data::DataDoc::element( 'website' );
   
   $ws->Attrib( 'id'    => $self->_get_id( 'website' ) );
   $ws->Attrib( 'root'  => $base );
   $self->{website}    = $ws;
   $self->{siteinfo}   = new XML::Data::DataDoc::element( 'siteinfo' );
   $ws->AddContent( $self->{siteinfo} );

   $self->{navigation} = new XML::Data::DataDoc::element( 'navigation' );
   $ws->AddContent( $self->{navigation} );

   $self->{resources}  = new XML::Data::DataDoc::element( 'resources' );
   $ws->AddContent( $self->{resources} );
  }


 ##--------------------------------
 #  parse the web site starting at the url given as a first argument into
 #    the current DataDoc.
 sub  parse_site
  {
   my $self = shift;

   $self->parse_url( $_[0] );

   print STDERR "Page info extracted from <$_[0]>.\n";
   die "No links found.\n" unless defined $self->{queue};

   while(@{$self->{queue}})
    {
     my ($url, $group, $set) = @{ shift @{$self->{queue}} };
     print STDERR "Extracting info from <$url>.\n";
  
     $self->{page} = $self->{pagelist}->{$url};

     $self->{group} = $group;
     $self->{set}   = $set;
     $self->parse_url( $url );
    }
  }


 ##--------------------------------
 #  handle the start tag for the <html> element
 sub  _html_start
  {
   my $self = shift;
   my $attr = shift;
   my $url  = $self->{url};
   my $full = _full_url( $url, $self->{base} );

   if(exists $self->{pagelist}->{$full})
    {
     $self->{page} = $self->{pagelist}->{$full};
    }
   else
    {
     $self->{page} = $self->_create_page( $url, $full );
     $self->_add_content( $self->{page} );
    }
  }


 ##--------------------------------
 #  create a page element in the current parser but don't add it to content.
 sub  _create_page
  {
   my $self = shift;
   my $url  = shift;
   my $full = shift || _full_url( $url, $self->{base} );

   my $page = new XML::Data::DataDoc::element( 'page' );

   $page->Attrib( 'href'  => $full );
   $page->Attrib( 'id'    => $self->_get_id( 'page' ) );
   $page->Attrib( 'title' => "untitled page" );

   $self->{pagelist}->{$full} = $page;
   $self->{get_content} = 1;

   $self->{website}->Attrib( 'main' => $page->Attrib( 'id' ) )
                               unless $self->{website}->HasAttrib( 'main' );

   $page;
  }


 ##--------------------------------
 #  handle the start tag for the <a> element
 sub  _a_start
  {
   my $self = shift;
   my $attr = shift;

   return unless defined $attr->{href};

   my $url = $attr->{href};
   $url =~ s/&/&amp;/g;

   if('mailto:' eq substr( $url, 0, length 'mailto:' ))
    {
      $url =~ s/^mailto://;
      unless(exists $self->{contactlist}->{$url})
       {
        my $curr = new XML::Data::DataDoc::element( 'contact' );

        $self->{curr} = $curr;
        $self->{siteinfo}->AddContent( $curr );

        $curr->Attrib( 'email' => $url );
        $curr->Attrib( 'id'    => $self->_get_id( 'contact' ) );
      
        $self->{get_content} = 1;
        $self->{contactlist}->{$url} = 1;
       }
    }
   elsif($url =~ /#/)
    {
     $self->{curr} = new XML::Data::DataDoc::element( 'subpage' );

     ($url, my $frag) = $url =~ /^([^#]*)#(.*)$/;

     $self->{curr}->Attrib( 'href'     => $url );
     $self->{curr}->Attrib( 'fragment' => $frag );
     $self->{get_content} = 1;
     $self->_add_content( $self->{curr} );
    }
   elsif($self->_url_on_site( $url ))
    {
     my $full = _full_url( $url, $self->{base} );
     if(exists $self->{pagelist}->{$full})
      {
       $self->{curr} = new XML::Data::DataDoc::element( 'pageref' );
       $self->{curr}->Attrib( 'ref' => $self->{pagelist}->{$full}->Attrib( 'id' ) );
    
       $self->{get_content} = 1;
       $self->_add_content( $self->{curr} );
      }
     elsif(lc $url !~ /\.(s?html?|asp|cfm)/ && $url !~ m#/$#)
      {
       my $curr = new XML::Data::DataDoc::element( 'file' );

       $curr->Attrib( 'href'  => _full_url( $attr->{href}, $self->{base} ) );
       $curr->Attrib( 'src'   => 'misc' );  # faked
       $curr->Attrib( 'dest'  => 'misc' );  # faked

       $self->_add_resource( $curr );
      }
     else
      {
       my $curr = $self->_create_page( $url, $full );
       $self->{curr} = $curr;

       if($url =~ m#/#)
        {
         my ($root) = $full =~ m#^(.*/)[^/]*$#;

         if(exists $self->{grouplist}->{$root})
          {
           $self->{group} = $self->{grouplist}->{$root};
          }
         else
          {
           my $grp = new XML::Data::DataDoc::element( 'group' );
           
           $self->_add_content( $grp );
           $grp->Attrib( 'root' => $root );
           $grp->Attrib( 'main' => $curr->Attrib( 'id' ) );
           $self->{grouplist}->{$root} = $grp;
           $self->{group} = $grp;
          }
         $curr->Attrib( 'href' => $full );

         if(exists $self->{setlist}->{$root})
          {
           $self->{set} = $self->{setlist}->{$root};
          }
         else
          {
           my $set = new XML::Data::DataDoc::element( 'set' );
           $self->_add_resource( $set );
           $self->{set} = $set;
           $set->Attrib( 'root' => $root );
           $self->{setlist}->{$root} = $self->{set};
          }

         $self->_add_content( $curr );
         push @{$self->{queue}}, [ $full, $self->{group}, $self->{set} ];
         delete $self->{group};
         delete $self->{set};
        }
       else
        {
         $self->_add_content( $curr );
         push @{$self->{queue}}, [ $full, $self->{group}, $self->{set} ];
        }
      }
    }
   else
    {
     my $curr = new XML::Data::DataDoc::element( 'fpage' );
     $self->{curr} = $curr;

     my $full = _full_url( $url, $self->{base} );
     unless(exists $self->{pagelist}->{$full})
      {
       $curr->Attrib( 'id'   => $self->_get_id( 'fpage' ) );
       $self->{pagelist}->{$full} = $curr;

       $curr->Attrib( 'href' => $url, 'title' => 'unnamed link' );
       $self->{get_content}  = 1;
       $self->_add_content( $curr );
      }
    }
  }

 ##--------------------------------
 #  handle the start tag for the <img> element
 sub  _img_start
  {
   my $self = shift;
   my $attr = shift;

   my $full = _full_url( $attr->{src}, $self->{base} );
   return if exists $self->{imglist}->{$full};

   $self->{imglist}->{$full} = 1;
   my $curr = new XML::Data::DataDoc::element( 'file' );

   $curr->Attrib( 'class' => 'image' );
   $curr->Attrib( 'href'  => $full );
   $curr->Attrib( 'src'   => 'images' );  # faked
   $curr->Attrib( 'dest'  => 'images' );  # faked

   $self->_add_resource( $curr );
  }



 ##--------------------------------
 #  handle the start tag for the <link> element
 sub  _link_start
  {
   my $self = shift;
   my $attr = shift;

   if(uc $attr->{rel} eq 'STYLESHEET')
    {
     my $full = _full_url( $attr->{href}, $self->{base} );

     if(exists $self->{stylelist}->{$full})
      {
       $self->{page}->Attrib( 'styleref' => $self->{stylelist}->{$full} );
      }
     else
      {
       my $curr = new XML::Data::DataDoc::element( 'stylesheet' );
       $curr->Attrib( 'id'   => $self->_get_id( 'stylesheet' ) );
       $curr->Attrib( 'type' => $attr->{type} || 'text/css' );
       $curr->Attrib( 'href' => $full );
       $self->{page}->Attrib( 'styleref' => $curr->Attrib( 'id' ) );
       $self->{stylelist}->{$full} = $curr->Attrib( 'id' );

       $self->{resources}->AddContent( $curr );
      }	
    }
  }

 ##--------------------------------
 #  handle the start tag for the <meta> element
 sub  _meta_start
  {
   my $self = shift;
   my $attr = shift;

   if(lc $attr->{name} eq 'keywords')
    {
     $self->{page}->Attrib( 'keywords' => $attr->{content} );
    }
   elsif(lc $attr->{name} eq 'description')
    {
     $self->{page}->Attrib( 'description' => $attr->{content} );
    }
  }


 ##--------------------------------
 #  handle the start tag for the <style> element
 sub  _style_start
  {
   my $self = shift;
   my $attr = shift;

   my $curr = new XML::Data::DataDoc::element( 'stylesheet' );
   $curr->Attrib( 'id'   => $self->_get_id( 'stylesheet' ) );
   $curr->Attrib( 'type' => $attr->{type} || 'text/css' );
   $curr->Attrib( 'href' => _full_url( $attr->{href}, $self->{base} ) )
                                            if defined $attr->{href};

   $self->{curr} = $curr;
   $self->{get_content} = 1;
   $self->{page}->Attrib( 'styleref' => $curr->{id} );

   $self->{resources}->AddContent( $curr );
  }


 ##--------------------------------
 #  handle the start tag for the <applet> element
 sub  _applet_start
  {
   my $self = shift;
   my $attr = shift;
   # ???  Need to check for duplicates

   my $curr = new XML::Data::DataDoc::element( 'file' );

   my $href = '';
   $href  = $attr->{codebase} ."/" if $attr->{codebase};
   $href .= $attr->{code};
   
   $curr->Attrib( 'class' => 'applet',
                  'src'   => 'applets',   # faked
                  'dest'  => 'applets',   # faked
                  'href'  => $href );

   $self->_add_resource( $curr );
  }



 ##--------------------------------
 #  handle start tags when parsing the html
 sub  start
  {
   my $self = shift;
   my $tag  = shift;
   my $attr = shift;

   if('html' eq $tag)
    {
     $self->_html_start( $attr );
    }
   elsif('a' eq $tag)
    {
     $self->_a_start( $attr );
    }
   elsif('img' eq $tag)
    {
     $self->_img_start( $attr );
    }
   elsif('link' eq $tag)
    {
     $self->_link_start( $attr );
    }
   elsif('meta' eq $tag)
    {
     $self->_meta_start( $attr );
    }
   elsif('title' eq $tag)
    {
     $self->{get_content} = 1;
    }
   elsif('style' eq $tag)
    {
     $self->_style_start( $attr );
    }
   elsif('applet' eq $tag)
    {
     $self->_applet_start( $attr );
    }
  }


 ##--------------------------------
 #  handle end tags when parsing the html
 sub  end
  {
   my $self = shift;
   my $tag  = shift;

   if('html' eq $tag)
    {
     delete $self->{page};
    }
   elsif('a' eq $tag and $self->{curr})
    {
     my $curr = $self->{curr};
     my $type = $curr->Type();
     $self->{get_content} = 0;

     if('page' eq $type or 'fpage' eq $type or 'subpage' eq $type)
      {
       my $title = _clean_attrib( $self->{text} || "unnamed link");

       $curr->Attrib( 'title' => $title || "unnamed link" );
      }
     elsif('contact' eq $type)
      {
       if($self->{text} =~ /\S/)
        {
         my $descr = new XML::Data::DataDoc::element( 'descr' );

         $self->{text} = _clean_attrib( $self->{text} );

         $descr->AddContent( new XML::Data::DataDoc::text( $self->{text} ) );
         $curr->AddContent( $descr );
        }
      }

     delete $self->{text};
     delete $self->{curr};
    }
   elsif('title' eq $tag)
    {
     $self->{get_content} = 0;
     $self->{text} = _clean_attrib( $self->{text} );
     $self->{page}->Attrib( 'title' => 
                            _clean_attrib( $self->{text} || "untitled page" ) );

     delete $self->{text};
    }
   elsif('style' eq $tag)
    {
     my $curr = $self->{curr};

     $self->{get_content} = 0;

     $curr->AddContent( new XML::Data::DataDoc::text( $self->{text} ) );

     delete $self->{text};
     delete $self->{curr};
    }
  }


 ##--------------------------------
 #  handle raw text when parsing the html
 sub  text
  {
   my $self = shift;

   $self->{text} .= shift if $self->{get_content};
  }


 ##--------------------------------
 #  handle comments when parsing the html
 sub  comment
  {
   my $self = shift;

   if($self->{get_content})
    {
     $self->{text} .= '&lt;!-- ';
     $self->{text} .= shift;
     $self->{text} .= ' --&gt;';
    }
  }


 ##--------------------------------
 #  Determine if the given url is on this site.
 sub  _url_on_site
  {
   my $self = shift;
   my $url = shift;

   return 0 unless exists $self->{base};

   my $full = _full_url( $url, $self->{base} );

   my $root = $self->{website}->Attrib( 'root' );
   $root eq substr($full->as_string, 0, length $root);
  }


 ##--------------------------------
 #  Add an element to the content
 sub  _add_content
  {
   my $self = shift;
   my $curr = shift;

   if($self->{group})
    {
     $self->{group}->AddContent( $curr );
    }
   else
    {
     $self->{navigation}->AddContent( $curr );
    }
  }

 ##--------------------------------
 #  Add an element to the resource list
 sub  _add_resource
  {
   my $self = shift;
   my $curr = shift;

   if($self->{set})
    {
     $self->{set}->AddContent( $curr );
    }
   else
    {
     $self->{resources}->AddContent( $curr );
    }
  }

 ##--------------------------------
 #  get full url given a relative url and a base
 sub  _full_url
  {
   my $url = new URI::URL( shift );

   $url->abs( shift );
  }


 ##--------------------------------
 #  get new id string
 sub _get_id
  {
   my $self = shift;
   my $el   = shift;
   
   $el . (++$self->{cntr}->{$el});
  }

##---------------------------------
#   clean an attribute string
sub _clean_attrib
 {
  my $s = shift;

  # remove leading/trailing whitespace
  $s =~ s/^\s+//;
  $s =~ s/\s+$//;

  # convert multiple spaces into single spaces
  $s =~ s/\s\s+/ /g;

  # handle invalid characters in attributes
  $s =~ s/&([^;]+)$/&amp;$1/;
  $s =~ s/&nbsp;//g;
  $s =~ s/>/&gt;/g;
  $s =~ s/</&lt;/g;
  $s =~ s/"/&quot;/g;
  # $s =~ s/'/&apos;/g;  # not for now.
 }


## ----- End of package -----
}


my $p = new HTML2DataDoc();

$p->parse_site( $ARGV[0] );

$p->{website}->Print();

