#!/usr/bin/perl -w
#
#  Copyright 2000-2002, G. Wade Johnson
#  Released under the same license as Perl.
#
#  Version: 0.7.5
#
#  Create a website using the supplied WSDL file.
#
use Getopt::Std;
use Data::Dumper;
use File::Copy;
use File::Path;
use File::Spec;
use File::Basename;
use URI::URL;
use FindBin;
use lib "$FindBin::Bin/lib";
use WSDL;
use WSDLtools;
use PLL;
use XML::Data::DataDoc;

use strict;
use vars qw/%options $bVerbose $bNonDestruct $IndexFile
            @Context %CmdLine @Ancestors %Macros %ToString/;

#--------------------------------------------
#  Define global variables and constants

use constant   DEFAULT_TARGET      => 'Site';
use constant   DEFAULT_SRC         => '.';
use constant   MACRO_NESTING_DEPTH => 10;

@Context   = ();
@Ancestors = ();
%CmdLine   = ();

my %CodeRefs  = ();
my @CodeEnds  = ();


# hash mapping resource elements to the routine used to copy them.
my %CopyResource =
 (
  stylesheet => \&copy_stylesheet,
  set        => \&copy_resource_set,
  page       => sub { },			# don't need to copy pages,
                                                # do I need to create it?
  file       => \&copy_resource_file,
  fpage      => sub { },			# don't need to copy fpages,
  subpage    => sub { },			# don't need to copy subpages,
 );

# hash mapping navigation elements to their creation routines
my %CreatePage =
 (
  page    => \&create_page,
  group   => \&create_group,
  website => \&create_website,
  file    => sub { },				# can't create files
  fpage   => sub { },				# can't create fpages
  subpage => sub { },				# can't create subpages
  pageref => sub { },				# can't create pagerefs
 );


# hash mapping macro names to the routines that implement them.
%Macros =
 (
  root         => sub { $Context[0]->{website}->Attrib( 'root' ) },
  source       => sub { $Context[0]->{website}->Attrib( 'src' ) },
  dest         => sub { $Context[0]->{website}->Attrib( 'dest' ) },
  footer       => sub { default_element_content_to_string( 'footer' ) },
  header       => sub { default_element_content_to_string( 'header' ) },
  content      => \&get_page_content,
  stylesheet   => \&get_stylesheet,
  keywords     => sub { meta_tag( 'keywords' ); },
  description  => sub { meta_tag( 'description' ); },
  full_url     => \&full_url,
  url_path     => \&url_path,
  body         => \&body_attributes,
  copyright    => \&copyright_text,
  # ??? need more macros
  # -- functions --
  timestamp    => \&timestamp,  # local unless arg = gmt.
  cmdline      => \&cmdline,
  # ??? need more macro functions
 );


# hash mapping element names to routines that return a string for that element.
%ToString =
 (
  server     => sub { $_[0]->Attrib( 'name' ) },
  directory  => sub { $_[0]->Attrib( 'name' ) },
  contact    => \&contact_to_string,
  stylesheet => \&stylesheet_to_string,
  footer     => \&element_content_to_string,
  header     => \&element_content_to_string,
  text       => \&element_content_to_string,
  # ??? what other elements do I need string conversions for?
 );


#--------------------------------------------
#  Begin main code

getopts( 'vns:d:i:', \%options ) or usage();

extract_defines();

my $file = $ARGV[0];

usage() unless $file;

# parse WSDL file
my $wsdl = new WSDL( $file, %CmdLine );

# validate the thing before we do any real work
$wsdl->validate();

my $website = $wsdl->website;

# initialize target directory information
$website->Attrib('dest' => $options{d} ) if defined $options{d};
$website->Attrib('src'  => $options{s} ) if defined $options{s};

$bVerbose     = $options{v};
$bNonDestruct = $options{n};
$IndexFile    = $options{i} || 'index.html';

# build site
prepare_code_pieces( $wsdl->IDedElementsByType( 'code' ) );
prepare_layouts( $wsdl );
init_context( $website );

create_navigatable_pages( NthContent( $website, 0, 'navigation' ) );
run_code_cleanup();
copy_site_resources( NthContent( $website, 0, 'resources' ) );

print "Site created successfully.\n";

#------------------------------------------------------------------
#  Subroutines.
#

sub  extract_defines
 {
  while(@ARGV)
   {
    last unless $ARGV[0] =~ /(\w+)=(.*)/;
    
    $CmdLine{$1} = $2;
    shift @ARGV;
   }
 }


#------------------------------------------------------------------
#  The NthContent function retrieves the 'nth' item from the content
#  of the specified element, needs the element, the value of n and
#  the same parameters as the Content() member function.
sub  NthContent
 {
  my $elem = shift;
  my $n    = shift || 0;

  ($elem->Content( @_ ))[$n];
 }



#------------------------------------------------------------------
#  The prepare_code_pieces function parses the code elements and
#  prepares them to execute at the appropriate times.
sub   prepare_code_pieces
 {
  foreach my $code (@_)
   {
    my $class = $code->Attrib( 'class' ) || 'routine';

    prepare_routine( $code ), next if $class eq 'routine';
    prepare_begin( $code ), next   if $class eq 'begin';
    prepare_end( $code ), next     if $class eq 'end';
    
    die "Code(".$code->Attrib( 'id' ).") has unrecognized class \'$class\'\n";
   }
 }


#------------------------------------------------------------------
#  The prepare_routine function creates a function out of the 'routine'
#  class of code elements and adds it to the %CodeRefs
sub   prepare_routine
 {
  my $code = shift;

  my $text = <<'EOC';
  sub {
   my $curr      = shift;
   my $ancestors = shift;
   my %args      = @_;
EOC
  my $lines = 5;

  $text .= $code->TextContent() . "\n }\n";

  my $id = $code->Attrib( 'id' );
  {
   local $SIG{__WARN__} = sub
    {
     my $m = shift;
     $m =~ s/at \(eval \d+\)/in code routine '$id'/g;
     $m =~ s/line (\d+)/"at line " . ($1 - $lines)/ge;
     warn $m;
    };
   
   $CodeRefs{$id} = eval $text;
  }

  if($@)
   {
     $@ =~ s/\(eval \d+\) line (\d+)/"line " . ($1 - $lines)/ge;
    die "Code($id) failed: $@\n" ;
   }

  print "code($id) compiled.\n" if $bVerbose;
 }


 
#------------------------------------------------------------------
#  The prepare_begin function prepares and executes any 'begin' class
#  code elements.
sub   prepare_begin
 {
  my $code = shift;
  my $id   = $code->Attrib( 'id' );

  {
   local $SIG{__WARN__} = sub
    {
     my $m = shift;
     $m =~ s/at \(eval \d+\)/in begin code '$id' at/g;
     warn $m; 
    };
   
   eval $code->TextContent();
  }

  if($@)
   {
    $@ =~ s/\(eval \d+\) //g;
    die "Code($id) failed: $@\n" ;
   }

  print "begin code($id) executed.\n" if $bVerbose;
 }


 
#------------------------------------------------------------------
#  The prepare_end function prepares any 'end' class code elements
#  for execution after the script completes.
sub   prepare_end
 {
  my $code = shift;
  my $id   = $code->Attrib( 'id' );

  {
   local $SIG{__WARN__} = sub
    {
     my $m = shift;
     $m =~ s/at \(eval \d+\)/in end code '$id'/g;
     $m =~ s/line (\d+)/at line $1/g;
     warn $m;
    };
   
    push @CodeEnds, [ $id, eval "sub { ". $code->TextContent() ." } " ];
   }

  print "end code($id) scheduled.\n" if $bVerbose;
 }


 
#------------------------------------------------------------------
#  The run_code_cleanup function executes all of the 'end' class
#  code elements prepared earlier.
sub   run_code_cleanup
 {
  foreach my $code (@CodeEnds)
   {
    eval {  $code->[1]->(); };
    
    die "End code($code->[0]) failed: $@" if $@;
    print "end code(", $code->Attrib('id'), ") executed.\n" if $bVerbose;
   }
 }




#------------------------------------------------------------------
#  Create all of the pages described by the page elements in the
#  navigation element.
sub   create_navigatable_pages
 {
  my  $nav = shift;

  return unless defined $nav;

  foreach my $np ($nav->ElementContent())
   {
    my $type = $np->Type();
    
    if(exists $CreatePage{$type})
     {
      $CreatePage{$type}->( $np );
     }
    else
     {
      warn "Unrecognized navigatable type \'$type'.\n";
     }
   }
 }



#------------------------------------------------------------------
#  Use the information from the page element to create an actual
#  HTML page.
sub   create_page( $np )
 {
  my $pg = shift;
  $Context[0]->{page} = $pg;

  set_page_properties( $pg );

  my $lo = $wsdl->byID( $pg->Attrib('layoutref') || $Context[0]->{layout} );
  
  die "No layout defined for page \'".$pg->Attrib('title')."\'." unless defined $lo;

  my ($destfile, $destpath);
  
  if($pg->HasAttrib('dest'))
   {
    $destfile = real_filename( 'dest', resolve_macros( $pg->Attrib('dest') ) );
    $destpath = dirname( $destfile );
   }
  else
   {
    $destpath = real_filename( 'dest', resolve_macros( $pg->Attrib('dest') ) );
    my $file  = basename( resolve_macros( $pg->Attrib('href') ), '' ) || $IndexFile;
    $destfile = buildfile( $destpath, $file );
   }

  print "Creating page \'".$pg->Attrib('title')."\' in \'$destfile\'.\n" if $bVerbose;  

  unless($bNonDestruct)
   {
    mkpath( $destpath ) unless !defined $destpath or "" eq $destpath or -d $destpath;

    open( OUT, ">$destfile" ) or die "Unable to create \'$destfile\'.\n";
  
    if('pll' eq $lo->Attrib('class'))
     {
      # walk pll building output string.
        # write file as needed
      print OUT PLL::convert( $lo->Attrib('data'), $pg, $wsdl, \@Context );
     }
    elsif('skeleton' eq $lo->Attrib('class'))
     {
      my $output = $lo->Attrib('data');

      print OUT resolve_macros( $output );
     }
    else
     {
      close( OUT );
      die "Unrecognized layout type \'".$lo->Attrib('class')."\'.\n";
     }
    close( OUT );
   }
   
  delete $Context[0]->{page};
  delete $Context[0]->{prop};
 }



sub   set_page_properties
 {
  my $pg = shift;

  $Context[0]->{prop} = 
    { map { $_->Attrib('name') => $_->Attrib('value') } $pg->Content( 'prop' ) };
 }




#------------------------------------------------------------------
#  Create a group of pages as described by the group element in the
#  navigation section of a website.
sub   create_group( $np )
 {
  my $grp = shift;
  
  print "Creating group ",
        (defined $grp->Attrib('id') ? $grp->Attrib('id') : 'unnamed'), ".\n"
                                                               if $bVerbose;
  push_website_context( $grp );
  unshift @Ancestors, $grp;

  # Handle children
  create_navigatable_pages( $grp );
  
  shift @Ancestors;
  pop_context();
 }




#------------------------------------------------------------------
#  Create a sub-site within the main site.
sub   create_website( $np )
 {
  my $ws = shift;
  
  print "Creating sub site ", $ws->Attrib('id'), ".\n" if $bVerbose;
  push_website_context( $ws );
  unshift @Ancestors, $ws;
  
  # Handle children
  create_navigatable_pages( NthContent( $ws, 0, 'navigation' ) );
  copy_site_resources( NthContent( $ws, 0, 'resources' ) );
  
  shift @Ancestors;
  pop_context();
 }



#------------------------------------------------------------------
#  Update the context for several defaults used in page creation.
sub   push_website_context
 {
  my $elem = shift;
  my $type = $elem->Type();
  my %ctx = ();
  my $tos = $Context[0];
  
  $ctx{layout}    = (resolve_macros( $elem->Attrib('layoutref') )
                     || $tos->{layout});
  $ctx{style}     = (resolve_macros( $elem->Attrib('styleref') )
                     || $tos->{style});
  $ctx{copyright} = (resolve_macros( $elem->Attrib('copyref') )
                     || $tos->{copyright});
  $ctx{keywords}  = (resolve_macros( $elem->Attrib('keywords') )
                     || $tos->{keywords});
  $ctx{description} = (resolve_macros( $elem->Attrib('description') )
                     || $tos->{description});
  $ctx{bodyref}   = (resolve_macros( $elem->Attrib('bodyref') )
                     || $tos->{bodyref});
  $ctx{group}     = 'group' eq $type   ? $elem : $tos->{group};
  $ctx{website}   = 'website' eq $type ? $elem : $tos->{website};

  $ctx{directory} = resolve_url( resolve_macros( $elem->Attrib('root') ),
                                       $tos->{directory} );
  $ctx{source}    = resolve_directory( resolve_macros( $elem->Attrib('src') ),
                                       $tos->{source} );
  $ctx{dest}      = resolve_directory( resolve_macros( $elem->Attrib('dest') ),
                                       $tos->{dest} );

  unshift @Context, \%ctx;
 }



#------------------------------------------------------------------
#  Restore the context to the previous nesting level
sub   pop_context
 {
  shift @Context;
 }


#------------------------------------------------------------------
#  Initialize the context array based on the current website's information.
sub  init_context
 {
  my $ws = shift;
  
  unshift @Context,
   { 
    dest      => resolve_macros( $ws->Attrib('dest') ) || DEFAULT_TARGET,
    source    => resolve_macros( $ws->Attrib('src') )  || DEFAULT_SRC,
    directory => resolve_macros( $ws->Attrib('root') ),
    layout    => resolve_macros( $ws->Attrib('layoutref') ),
    style     => resolve_macros( $ws->Attrib('styleref') ),
    copyright => resolve_macros( $ws->Attrib('copyref') ),
    keywords  => resolve_macros( $ws->Attrib('keywords') ),
    description => resolve_macros( $ws->Attrib('description') ),
    bodyref   => $ws->Attrib('bodyref'),
    website   => $ws,
   };

  unshift @Ancestors, $ws;
 }



#------------------------------------------------------------------
#  Replace macros in a string with their appropriate values.
sub   resolve_macros
 {
  my $s     = shift;
  my $depth = shift || 0;

  return $s unless $s;

  die "Macro nesting is too deep.\n" if $depth > MACRO_NESTING_DEPTH;

  if($s =~ m/{{(([^}]|}[^}])+)}}/)
   {
    $s =~ s/{{(([^}]|}[^}])+)}}/resolve_macro($1, $depth+1)/smeg;
   }

  $s;
 }


#------------------------------------------------------------------
#  Handle nested macros.
sub   resolve_macro
 {
  my $s     = shift;
  my $depth = shift;
  
  resolve_macros( _resolve_macro( $s, $depth ) || '' );
 }


#------------------------------------------------------------------
#  Replace a macro in a string with its appropriate values.
sub  _resolve_macro
 {
  my $match = shift;

    # defined simple macro
  if($match =~ m/^(\w+)(?:\(([^)]*)\))?$/)
   {
    return resolve_builtin_macro( $1, $2 );
   }
        # code routine
  elsif($match =~ m#^code\[(\w+)\]\(([^)]*)\)#)
   {
    my ($id,$args) = ($1, $2);
    return exec_code_element( $id, ($args =~ m#\s*(\S+)\s*=\s*(\S*)#g) );
   }
        # element attribute
  elsif($match =~ m/^\@(\w+)/)
   {
    return $Context[0]->{page}->Attrib($1);
   }
        # properties of this page
  elsif($match =~ m/^prop\[(\w+)\]$/)
   {
    return resolve_macros( ($Context[0]->{prop}->{$1}) || '' );
   }
        # particular element value
  elsif($match =~ m/^(\w+)\[(\w+)\](\/.*)?/)
   {
    return resolve_particular_element( $1, $2, $3 );
   }
  else
   {
    die "Unrecognized syntax: \'$match\'.\n";
   }
 }


#--------------------------------------------------------------
#  Resolve a builtin macro with optional arguments.
sub   resolve_builtin_macro
 {
  my $macro = shift;
  
  die "Unrecognized macro \'$macro\'.\n" unless exists $Macros{$macro};
  
  return $Macros{$macro}->( shift );
 }



#--------------------------------------------------------------
#  Convert an element (with optional attribute) into a string
#  for display.
sub   resolve_particular_element
 {
  my ($type,$id,$rest) = @_;
  my $elem = $wsdl->byID( $id );

  die "No $type named $id.\n" if !defined $elem or $type ne $elem->Type();

  unless($rest)
   {
    return $elem unless exists $ToString{$type};

    return $ToString{$type}->( $elem );
   }
  elsif($rest =~ m#^/@(\w+)$#)
   {
    return $elem->Attrib( $1 );
   }
  else
   {
    die "Unable to parse \'$rest\'.\n";
   }
 }


#------------------------------------------------------------------
#  Update the context for the defaults for resources.
sub   push_resource_context
 {
  my $elem = shift;
  my %ctx = ();
  my $tos = $Context[0];
  my $dest = $elem->Attrib('dest');

  if(!defined $elem->Attrib('dest') and defined $elem->Attrib('root'))
   {
    my $root = new URI::URL( $elem->Attrib('root') );
    $dest = $root->path;
   }

  $ctx{directory} = resolve_url( resolve_macros( $elem->Attrib('root') ),
                                 $tos->{directory} );
  $ctx{source}    = resolve_directory( resolve_macros( $elem->Attrib('src') ),
                                       $tos->{source} );
  $ctx{dest}      = resolve_directory( resolve_macros( $dest ),
                                       $tos->{dest} );
  $ctx{website}   = $tos->{website};

  unshift @Context, \%ctx;
 }





#------------------------------------------------------------------
#  Copy the site resources to the appropriate target locations.
sub   copy_site_resources
 {
  my  $res = shift;

  return unless defined $res;

  push_resource_context( $res );

  foreach my $f ($res->ElementContent())
   {
    my $type = $f->Type();

    if(exists $CopyResource{$type})
     {
      $CopyResource{$type}->( $f );
     }
    else
     {
      warn "Unrecognized resource type \'$type'.\n";
     }
   }

  pop_context();
 }



#------------------------------------------------------------------
#  Copy a stylesheet to the target location.
sub   copy_stylesheet
 {
  my $ss = shift;

  copy_resource_file( $ss, 'copy stylesheet' ) if $ss->HasAttrib('href');
 }



#------------------------------------------------------------------
#  Copy a set of resources to the appropriate locations.
sub   copy_resource_set
 {
  my $set = shift;
  my $id  = $set->Attrib('id') || "unknown";

  push_resource_context( $set );
  unshift @Ancestors, $set;

  print "Copy resource set \'$id\'.\n" if $bVerbose;

  foreach my $f ($set->ElementContent())
   {
    my $type = $f->Type();

    if(exists $CopyResource{$type})
     {
      $CopyResource{$type}->( $f );
     }
    else
     {
      warn "Unrecognized resource type \'$type'.\n";
     }
   }

  shift @Ancestors;
  pop_context();
 }



#------------------------------------------------------------------
#  Copy an individual resource file to the correct location.
sub   copy_resource_file
  {
   my $f  = shift;
   my $msg = shift || 'copy resource';

   my $destfile = real_filename( 'dest', $f->Attrib('dest') );
   my $destpath = dirname( $destfile );
   my $srcfile  = real_filename( 'source', $f->Attrib('src') );
   my $file     = basename( resolve_macros( $f->Attrib('href') ), '' );
   $destfile    = buildfile( $destpath, $file ) if $destfile eq $destpath;

   $srcfile = buildfile( $srcfile, $file ) unless -f $srcfile;

   unless($bNonDestruct)
    {
     mkpath( $destpath ) unless !defined $destpath or "" eq $destpath or -d $destpath;
     die "File \"$srcfile\" not found.\n" unless -f $srcfile;
     copy( $srcfile, $destfile );
    }
   else
    {
     warn "File \"$srcfile\" not found.\n" unless -f $srcfile;
    }

   print "$msg $srcfile => $destfile\n" if $bVerbose;
  }



#------------------------------------------------------------------
#  Create a local path out of the supplied directory components
sub   buildpath
 {
  my @dirs = grep { defined $_ and $_ ne "" } @_;

  return "" unless @dirs;

  my $dir = shift @dirs;

  while(@dirs)
   {
    $dir = File::Spec->file_name_is_absolute( $dirs[0] )
             ? shift @dirs : File::Spec->catdir( $dir, shift @dirs );
   }

  $dir;
 }


#------------------------------------------------------------------
#  Create a path to a local file out of the supplied directory
#  and file components
sub   buildfile
 {
  my @dirs = grep { defined $_ and $_ ne "" } @_;

  return "" unless @dirs;

  my $file = pop @dirs;
  return $file unless @dirs;

  my $dir = shift @dirs;

  while(@dirs)
   {
    $dir = File::Spec->file_name_is_absolute( $dirs[0] )
             ? shift @dirs : File::Spec->catdir( $dir, shift @dirs );
   }

  File::Spec->catfile( $dir, $file );
 }


#------------------------------------------------------------------
#  Generate the actual filename with path, given the name of the context
#  directory and the filename.
sub   real_filename
 {
  my $dirtype = shift;
  my $file    = shift;

  return $Context[0]->{$dirtype} unless defined $file;

  if(!$Context[0]->{$dirtype} or File::Spec->file_name_is_absolute( $file ))
   {
    $file;   # return the filename if it is absolute
   }
  else
   {
    buildpath( $Context[0]->{$dirtype}, $file );
   }
 }




#------------------------------------------------------------------
#  Concatenate directories as appropriate.
sub   resolve_directory
 {
  my $dir  = shift;
  my $base = shift;

  if(defined $dir and defined $base and !File::Spec->file_name_is_absolute( $dir ))
   {
    $dir = File::Spec->catdir( $base, $dir );
   }
  elsif(!defined $dir and defined $base)
   {
    $dir = $base;
   }

  $dir = File::Spec->canonpath( $dir )  if defined $dir;
  $dir;
 }




#------------------------------------------------------------------
#  Concatenate url directories as appropriate.
sub   resolve_url
 {
  my $rel  = shift;
  my $base = shift;
  my $bAbs = shift;

  $bAbs = ($website->HasAttrib('linktype')
           and 'absolute' eq $website->Attrib('linktype'))
                                                  unless defined $bAbs;

  $rel = resolve_macros( $rel );

  if(defined $rel and $rel ne '' and defined $base)
   {
    my $url = new URI::URL( $rel );

    $rel = $url->abs( $base );
   }
  elsif(defined $base)
   {
    $rel = $base;
   }

  if($bAbs)
   {
    $rel;
   }
  else
   {
    my $url = new URI::URL( $rel );
    $url->full_path();
   }
 }


#------------------------------------------------------------------
#  Read in the layout files and prepare them for use.
sub   prepare_layouts
 {
  my $wsdl = shift;
  my $p = new XML::Parser( Style=>'XML::Data::DataDoc::Style' );

  foreach my $lo ($wsdl->IDedElementsByType( 'layout' ))
   {
    my $id    = $lo->Attrib('id');
    my $class = $lo->Attrib('class') || 'pll';
    my $file  = resolve_macros( $lo->Attrib('file') );

    if('pll' eq $class)
     {
      $lo->Attrib('data' => ($p->parsefile( $file ))->Root() );

      PLL::validate( $lo->Attrib('data'), $wsdl );
      print "Layout \'", $lo->Attrib('id'), "\' is valid.\n" if $bVerbose;
     }
    elsif('skeleton' eq $class)
     {
      local $/ = undef;
      local *F;

      open( F, "$file" ) or die "Unable to open layout file \'$file\': $!\n";
      my $file = <F>;
      $lo->Attrib( 'data'  => $file );
      close( F );
     }
    else
     {
      die "Unrecognized layout class \'$class\'.\n";
     }
   }
 }



#------------------------------------------------------------------
#  Convert a contact to a string
sub   contact_to_string
 {
  my $contact = shift;

  join( ' ', map { ($_->HasAttrib('class') and $_->Attrib('class') eq 'nickname') ? 
                      '"'.$_->TextContent().'"' : $_->TextContent() } 
             $contact->ElementContent()
      );
# ??? expand to handle name parts better.
 }



#------------------------------------------------------------------
#  Convert a stylesheet to a string
sub   stylesheet_to_string
 {
  my $ss = shift;
  my $type = $ss->Attrib('type') || "text/css";

  if($ss->HasAttrib('href'))
   {
    my $href = resolve_macros( $ss->Attrib('href') );
    my $url = resolve_url( $href, $Context[0]->{directory} );

    qq{<link rel="STYLESHEET" type="$type" href="$url">\n};
   }
  else
   {
    qq{<style type="$type">\n} . $ss->TextContent() . qq{</style>\n};
   }
 }



#------------------------------------------------------------------
#  return the text content of the element (after macro expansion)
sub    element_content_to_string
 {
  my $elem = shift;

  resolve_macros( $elem->TextContent() );
 }



#------------------------------------------------------------------
#  Convert the specified body element into attributes for a body
#  tag.
sub    body_attributes
 {
  my $id     = shift || $Context[0]->{page}->Attrib('bodyref')
                     || $Context[0]->{bodyref};  
  return '' unless defined $id and '' ne $id;
  
  my $elem   = $wsdl->byID( $id );
  die "Unrecognized body \'$id\'.\n" unless defined $elem and 'body' eq $elem->Type();
  
  my $output = '';

  my @aNames = $elem->AttribNames();

  if(@aNames)
   {
    $output .= join( ' ', map { qq{$_="}.$elem->Attrib($_).qq{"} } 
                          grep { $_ ne 'id' }
                          @aNames );
   }

  resolve_macros( $output );
 }



sub   copyright_text
 {
  my $id     = shift || $Context[0]->{page}->Attrib('copyref')
                     || $Context[0]->{copyright};  
  return '' unless defined $id and '' ne $id;
  
  my $elem   = $wsdl->byID( $id );
  die "Unrecognized copyright \'$id\'.\n" unless defined $elem
                                             and 'copyright' eq $elem->Type();
  
  my $year = $elem->Attrib('year') || (localtime)[5];
  my @wholist = split( '\s+', $elem->Attrib('owner') || '' );
  my $name = undef;

  if(@wholist == 1)
   {
    $name = copyight_owner_name( $wholist[0] );
   }
  elsif(@wholist == 2)
   {
    $name  = copyight_owner_name( $wholist[0] );
    $name .= " and " . copyight_owner_name( $wholist[1] );
   }
  elsif(@wholist)
   {
    @wholist = map {copyight_owner_name( $_ ) } @wholist;
    my $last = pop @wholist;
    $name = join( ', ', @wholist, "and $last" );
   }

  if(exists $elem->{style} and $elem->{style} eq 'copy')
   {  
    join( ' ', 'Copyright &copy;', $year, $name );
   }
  else
   {  
    join( ' ', 'Copyright', $year, $name );
   }
 }


sub  copyight_owner_name
 {
  my $who = shift;
  my $contact = $wsdl->byID( $who );
  die "Unrecognized contact \'$who\'.\n" unless defined $contact
                                       and 'contact' eq $contact->Type();

  contact_to_string( $contact );
 }



#------------------------------------------------------------------
#  Build meta tags from appropriate attributes
sub   meta_tag
 {
  my $attrib = shift;
  my $pg     = $Context[0]->{page};
  my $output = $pg->Attrib($attrib) || $Context[0]->{$attrib};

  $output ? qq{<meta name="$attrib" content="$output">\n} : $output;
 }



#------------------------------------------------------------------
#  return the text content of the element (after macro expansion)
sub    default_element_content_to_string
 {
  my $type = shift;

  element_content_to_string( ($wsdl->IDedElementsByType( $type ))[0] );
 }



#------------------------------------------------------------------
#  Execute the code associated with the named code element
sub   exec_code_element
 {
  my $id = shift;

  die "Unrecognized code reference \'$id\'.\n" unless exists $CodeRefs{$id};

  $CodeRefs{$id}->( $Context[0]->{page}, \@Ancestors, @_ );
 }




#------------------------------------------------------------------
#  Execute the code associated with the named code element
sub   get_page_content
 {
  my $pg = $Context[0]->{page};

  return "" unless $pg->HasAttrib('content');

  my $file = buildfile( $Context[0]->{source}, $pg->Attrib('src'), 
                        resolve_macros( $pg->Attrib('content') ) );
  my $output ="";

   {
    local *F;
    local $/ = undef;

    open( F, $file ) or die "Unable to open \'$file\'.\n";
    $output = <F>;
    close( F );

    # Extract body of the HTML
    $output =~ s#^.*<[Bb][Oo][Dd][Yy][^>]*>##s;
    $output =~ s#</[Bb][Oo][Dd][Yy]>.*$##s;

    $output = resolve_macros( $output );
   }

  $output;
 }



#------------------------------------------------------------------
#  Track down the stylesheets associated with this page and generate
#   output to include it.
sub   get_stylesheet
 {
  my $pg = $Context[0]->{page};
  my $ssid = $pg->Attrib('styleref') || $Context[0]->{style};

  return "" unless $ssid;

  my @stylesheets = ();
  foreach my $ssref (split( '\s+', $ssid ))
   {
    my $ss = $wsdl->byID( $ssref );
    die "Unable to find stylesheet \'$ssref\'.\n"
                     if !defined $ss or 'stylesheet' ne $ss->Type();

    push @stylesheets, stylesheet_to_string( $ss );
   }

  join( '', @stylesheets ); 
 }


#------------------------------------------------------------------
#  Generate a string giving the full url of the current page.
sub   full_url
 {
  my $href = resolve_macros( $Context[0]->{page}->Attrib('href') );
  my $url = resolve_url( $href, $Context[0]->{directory}, 1 );
  $url;
 }


#------------------------------------------------------------------
#  Generate a string giving the relative url of the current page.
sub  url_path
 {
  my $href = resolve_macros( $Context[0]->{page}->Attrib('href') );
  my $url = resolve_url( $href, $Context[0]->{directory}, 0 );
  $url;
 }



#------------------------------------------------------------------
#  Generate a string giving the current time in local time unless
#  the argument 'gmt' is supplied, then return time in gmt.
sub   timestamp
 {
  my $arg = shift;

  $arg =~ s/^\s+//, $arg =~ s/\s+$// if defined $arg;

  if(defined $arg and 'gmt' eq lc $arg)
   {
    scalar gmtime time;
   }
  else
   {
    scalar localtime time;
   }
 }

 
#------------------------------------------------------------------
#  Return the value of the requested command line parameter as a
#   string.
sub   cmdline
 {
  my $arg = shift;

  $arg =~ s/^\s+//, $arg =~ s/\s+$// if defined $arg;

  if(defined $arg and exists $CmdLine{$arg})
   {
    $CmdLine{$arg};
   }
  else
   {
    '';
   }
 }

 
#------------------------------------------------------
#  Expects elements, not strings
# ???? need to reverify
sub   make_href
 {
  my $page  = shift;

  my $href = resolve_macros( $page->Attrib('href') ) if $page->HasAttrib('href');

  if(!defined $href or relative_url( $href ))
   {
    foreach my $a (@_)
     {
      next unless defined $a;

      unless(defined $href)
       {
        $href = resolve_macros( $a->Attrib('root') ) if $a->HasAttrib('root');
        next;
       }

      $href = resolve_url( $href, resolve_macros( $a->Attrib('root') ) );
      last unless relative_url( $href );
     }
   }

  $href || "";
 }


#-----------------------------------------------------
#   Is the supplied URL a relative link?
sub  relative_url
 {
  my $url = shift;

  $url !~ /^http:/ and $url !~ m#^/#;
 }




#------------------------------------------------------------------
#  Explain usage of the program.
sub   usage
 {
  die <<"USAGE"
Usage:  $0  [-v] [-n] [-d target] [-s source] [-i indexfile] wsdlfile

where:  -v            - verbose output
        -n            - don't copy files
        -d target     - the target directory for writing the website files.
        -s source     - the source directory for web site portions
        -i indexfile  - set the name of the directory index file
        wsdlfile      - the WSDL description of the website to build.

USAGE
 }
