#!/usr/bin/perl
#
#  Copyright 2000-2002, G. Wade Johnson
#  Released under the same license as Perl.
#  
#  Version: 0.7.6
#
#  This module implements a routine to create a WSDL data structure
#  from an existing file.
#

package WSDL;

use File::Basename;
use File::Spec;
use XML::Data::DataDoc;
use XML::Data::DataDoc::Style;
use Data::Dumper;
use strict;

##---------------------------------------------------------------
# Data structures used in creation
my %IncludeAttribs = ();  # filled below

my %ContainText =
   map { $_ => 1 } (qw/descr honorific name degree institution copyright text
                      header footer stylesheet code/,
		      XML::Data::DataDoc::xmltypeCData
		   );

##---------------------------------------------------------------
# Data structures used in validation
my %Any = ();

my %LegalTags =
 (
  body        => { attr     => { id=> 0, class => 0, style => 0, title => 0,
                                 lang => 0, dir => 0, onclick => 0,
				 ondblclick => 0, onmousedown => 0,
				 onmouseup => 0, onmouseover => 0,
                                 onmousemove => 0, onmouseout => 0,
				 onkeypress => 0, onkeydown => 0, onkeyup => 0,
				 onload => 0, onunload => 0, background => 0,
				 bgcolor => 0, text => 0, link => 0, vlink => 0,
				 alink => 0,
                                },
                   required => [ 'id' ],
                   content  => {},
                 },
  choose      => { attr     => {},
                   required => [],
                   content  => { when => 1, otherwise => 0 },
                 },
  code        => { attr     => { id => 0, include => 0, class => 0 },
                   required => [ 'id' ],
                   content  => {},
                 },
  contact     => { attr     => { id => 0, href => 0, email => 0 },
                   required => [ 'id' ],
                   content  => { descr => 0, honorific => 0, name => 0,
		                 degree => 0, institution => 0
                               },
                 },
  copyright   => { attr     => { id => 0, year => 0, owner => 0, style => 0 },
                   required => [ 'id' ],
                   content  => {},
                 },
  degree      => { attr     => {},
                   required => [],
                   content  => {},
                 },
  descr       => { attr     => {},
                   required => [],
                   content  => {},
                 },
  directory   => { attr     => { id => 0, name => 0, class => 0 },
                   required => [ 'id', 'name' ],
                   content  => {},
                 },
  file        => { attr     => { id => 0, src => 0, dest => 0, href => 0,
                                 class => 0, type => 0, nav => 0, title => 0
			       },
                   required => [ 'href' ],
                   content  => {},
                 },
  footer      => { attr     => { id => 0, include => 0 },
                   required => [ 'id' ],
                   content  => {},
                 },
  fpage       => { attr     => { title => 0, href => 0, fragment => 0, id => 0 },
                   required => [ 'title', 'href' ],
                   content  => {},
                 },
  group       => { attr     => { id => 0, root => 0,  main => 'page', href => 0,
                                 include => 0, title => 0,
				 copyref => 'copyright', keywords => 0,
				 description => 0, styleref => 0,
				 layoutref => 'layout', src => 0, dest => 0
			        },
                   required => [],
                   content  => { group => 0, page => 0,  pageref => 0,
		                 subpage => 0, fpage => 0, file => 0,
				 website => 0, 'if' => 0, choose => 0
                               },
                 },
  header      => { attr     => { id => 0, include => 0 },
                   required => [ 'id' ],
                   content  => {},
                 },
  honorific   => { attr     => {},
                   required => [],
                   content  => {},
                 },
  'if'        => { attr     => { test => 0 },
                   required => [ 'test' ],
                   content  => \%Any,
                 },
  institution => { attr     => {},
                   required => [],
                   content  => {},
                 },
  layout      => { attr     => { id => 0, file => 0, class => 0 },
                   required => [ 'id', 'file' ],
                   content  => {},
                 },
  name        => { attr     => { class => 0 },
                   required => [],
                   content  => {},
                 },
  navigation  => { attr     => { id => 0, include => 0 },
                   required => [],
                   content  => { group => 0, page => 0,  pageref => 0,
		                 subpage => 0, fpage => 0, file => 0,
				 website => 0, 'if' => 0, choose => 0
                               },
                 },
  otherwise   => { attr     => {},
                   required => [],
                   content  => \%Any,
                 },
  page        => { attr     => { title => 0, src => 0, dest => 0, href => 0,
                                 id => 0, keywords => 0, description => 0,
				 skel => 0, content => 0, styleref => 0,
				 copyref => 0, bodyref => 'body',
				 layoutref => 'layout'
			       },
                   required => [ 'title' ],
                   content  => { prop => 0, 'if' => 0, choose => 0 },
                 },
  pageref     => { attr     => { ref => 0, fragment => 0, title => 0 },
                   required => [ 'ref' ],
                   content  => {},
                 },
  prop        => { attr     => { name => 0, value=> 0 },
                   required => [ 'name', 'value' ],
                   content  => {},
                 },
  resources   => { attr     => { include => 0 },
                   required => [],
                   content  => { stylesheet => 0, set => 0, page => 0, 
                                 file => 0, fpage => 0, subpage => 0,
                                 'if' => 0, choose => 0
                               },
                 },
  server      => { attr     => { id => 0, name => 0, class => 0 },
                   required => [ 'id', 'name' ],
                   content  => {},
                 },
  set         => { attr     => { id => 0, include => 0, root => 0, src => 0,
                                 dest => 0
			       },
                   required => [],
                   content  => { stylesheet => 0, set => 0, page => 0, 
                                 file => 0, fpage => 0, subpage => 0,
                                 'if' => 0, choose => 0
                               },
                 },
  siteinfo    => { attr     => { id => 0, include => 0, ref => 'page' },
                   required => [],
                   content  => { server => 0,  directory => 0,  contact => 0, 
                                 copyright => 0, text => 0, header => 0, 
                                 footer => 0, layout => 0, body => 0, 'if' => 0,
                                 choose => 0,
                               },
                 },
  subpage     => { attr     => { title => 0, href => 0, fragment => 0, id => 0 },
                   required => [ 'title', 'fragment' ],
                   content  => {},
                 },
  stylesheet  => { attr     => { id => 0, type => 0, href => 0, src => 0,
                                 dest => 0
			       },
                   required => [ 'id', 'type' ],
                   content  => {},
                 },
  text        => { attr     => { id => 0, class => 0, include => 0 },
                   required => [ 'id' ],
                   content  => {},
                 },
  website     => { attr     => { id  => 0, include => 0, src =>0, dest => 0,
                                 root => 0, main => 'page', title => 0,
                                 keywords => 0, description => 0,
				 styleref => 'stylesheet', linktype => 0,
				 copyref => 'copyright', bodyref => 'body',
                                 layoutref => 'layout'
				},
                   required => [ 'id', 'main', 'root' ],
                   content  => { siteinfo => 0, navigation => 1, resources => 0,
                                 code => 0
			       },
                 },
  when        => { attr     => { test => 0 },
                   required => [ 'test' ],
                   content  => \%Any,
                 },
 );

%Any = map { $_ => 0 } keys %LegalTags;

%IncludeAttribs = map  { $_ => 1 }
                  grep { exists $LegalTags{$_}->{attr}->{include} }
                  keys %LegalTags;

##------------------------------------------------------------------------
##  The following package generates a hash of ids to elements while it
##  builds the standard DataDoc object. This would be more general, if I
##  could figure out a good way to pass the information needed to recognize
##  which elements have ID attributes and what the names of those attributes
##  are.
{
 package WSDL::IDedObject;
 @WSDL::IDedObject::ISA = ( 'XML::Data::DataDoc::Style' );
 use vars( '%CmdLine' );

 %CmdLine = ();

 sub Init
  {
   my $expat = shift;
   
   XML::Data::DataDoc::Style::Init( $expat );

   $expat->{IDs} = {};
  }
  
 sub XMLDecl { &XML::Data::DataDoc::Style::XMLDecl; } 

 sub  _ignore
  {
   my $expat = shift;

   $expat->{ignore}->[0] or $expat->{cond}->[0];
  }


 sub  Start
  {
   my $expat = shift;
   my $type = shift;
   my $elem = new XML::Data::DataDoc::element( $type );
 
   unless(_ignore( $expat ))
    {
     if('if' eq $type)
      {
       unshift @{$expat->{ignore}},
                 _test_conditional( $expat, $_[1] ) ? undef : $elem;
      }
     elsif('choose' eq $type)
      {
       unshift @{$expat->{choose}}, 0;
      }
     elsif('when' eq $type)
      {
       if($expat->{choose}->[0])
        {
         unshift @{$expat->{ignore}}, $elem;
        }
       else
        {
         unshift @{$expat->{ignore}},
                   _test_conditional( $expat, $_[1] ) ? undef : $elem;
        }       
      }
     elsif('otherwise' eq $type)
      {
       unshift @{$expat->{ignore}}, ($expat->{choose}->[0] ? $elem : undef);
      }
     else
      {
       while(@_)
        { 
         my $n = shift; 
         $elem->Attrib( $n => shift );

         if($n eq 'id')
          {
           die $elem->Type() . " duplicates id " . $elem->Attrib( $n ) ."\n" 
                                 if exists $expat->{IDs}->{$elem->Attrib($n)};
           $expat->{IDs}->{$elem->Attrib($n)} = $elem;
          }
        }
      }
    }
 
   push @{$expat->{Partial}}, $elem; 
  }


 sub _test_conditional
  {
   my $expat = shift;
   my $test  = shift;
   my $result;

   eval "\$result = $test";

   $result;
  }



 sub End
  {
   my $expat = shift;
   my $tag = shift;

   my $elem = pop @{$expat->{Partial}};
   die "Improper element nesting.\n" if $tag ne $elem->Type();

   if(_ignore( $expat ))
    {    
     if($elem == $expat->{ignore}->[0])
      {
       if('if' eq $tag or 'when' eq $tag or 'otherwise' eq $tag)
        {
         shift @{$expat->{ignore}};
        }
      }
    }
   else
    {
     if('if' eq $tag)
      {
       foreach($elem->Content())
        {
         _insert_content( $expat, $_ );
        }
       shift @{$expat->{ignore}};
      }
     elsif('when' eq $tag or 'otherwise' eq $tag)
      {
       shift @{$expat->{ignore}};
       _insert_content( $expat, $elem );
       $expat->{choose}->[0] = 1;
      }
     elsif('choose' eq $tag)
      {
       my $success = ($elem->Content('*'))[0];

       foreach($success->Content())
        {
         _insert_content( $expat, $_ );
        }

       shift @{$expat->{ignore}};
       shift @{$expat->{choose}};
      }
     else
      {
       _insert_content( $expat, $elem );
      }
    }
  }


 sub  _insert_content
  {
   my $expat = shift;
   my $elem  = shift;

   if(@{$expat->{Partial}})
    {
     $expat->{Partial}->[-1]->AddContent( $elem );
    }
   else
    {
     $expat->{Tree}->AddContent( $elem );
    }
  }

 sub Char
  {
   my $expat = $_[0];
   return if _ignore( $expat );
   
   if(exists $ContainText{$expat->{Partial}->[-1]->Type()})
    {
     &XML::Data::DataDoc::Style::Char;
    }
   elsif($_[1] !~ /^\s*$/)
    {
     die "CDATA content not allowed in ", 
         $expat->{Partial}->[-1]->Type(), " element.\n";
    }
  }
 sub Comment { }   # Dump comments
 sub Proc
  {
   return if _ignore( $_[0] );
   &XML::Data::DataDoc::Style::Proc;
  }
 sub Final
  {
   my $expat = shift;
   
   my $doc = XML::Data::DataDoc::Style::Final( $expat );
   
   ($doc, $expat->{IDs});
  }
 sub CdataStart
  {
   return if _ignore( $_[0] );
   &XML::Data::DataDoc::Style::CdataStart;
  }
 sub CdataEnd
  {
   return if _ignore( $_[0] );
   &XML::Data::DataDoc::Style::CdataEnd;
  }
}

# --------------------------------------------------------------
#  Create a WSDL object
sub   new
 {
  my $pkg  = shift;
  my $file = shift;
  my $self = { doc => {},
               IDs => {}
             };
  bless $self, $pkg;

  %WSDL::IDedObject::CmdLine = @_;
  
  $self->{parser} = new XML::Parser( Style=>'WSDL::IDedObject' );

  $self->parsefile( $file ) if defined $file;

  $self;  
 }

# ---------------------------------------------------------------
#  Use the XML::Parser created in 'new' to parse a file into the
#  current WSDL object.
sub  parsefile
 {
  my $self = shift;
  my $file = shift;

  $self->{basedir} = dirname( $file );
  
  ($self->{doc}, $self->{IDs}) = $self->{parser}->parsefile( $file );

  # Now I need to walk the element list and handle any includes
  $self->_resolve_includes( $self->{doc}->Root() );
 }


# ---------------------------------------------------------------
#  Use the XML::Parser created in 'new' to parse a string into the
#  current WSDL object.
sub  parse
 {
  my $self = shift;
  my $str = shift;

  $self->{basedir} = '.';
  
  ($self->{doc}, $self->{IDs}) = $self->{parser}->parse( $str );

  # Now I need to walk the element list and handle any includes
  $self->_resolve_includes( $self->{doc}->Root() );
 }


##----------------------------------------------------------------
##  Finds elements keyed in %IncludeAttribs above, checks to see if
##  they have the include attribute, if so, include the file and
##  recurse
sub   _resolve_includes
 {
  my  $self = shift;
  my  $root = shift;

  foreach my $e ($root->Content( '*' ))
   {
    $self->_resolve_includes( $e ) unless $e->IsEmpty;
    
    #   1. find each element with an include.
    next unless exists $IncludeAttribs{$e->Type()} and $e->HasAttrib( 'include' );

                       # Remove the current ID
    delete $self->{IDs}->{$e->Attrib('id')} if $e->HasAttrib( 'id' );

    #   2. run parser on the file that is found.
    my $file = $e->Attrib('include');
    $file = File::Spec->catfile( $self->{basedir}, $file )
                 unless -f $file or File::Spec->file_name_is_absolute( $file );

    my ($doc,$ids) = $self->{parser}->parsefile( $file );
    die "Unable to include \'$file\'.\n" unless defined $doc;
    
    if(defined $ids)
     {
      foreach my $key (keys %{$ids})
       {
        $self->{IDs}->{$key} = $ids->{$key};
       }
     }

    #   3. take the root of the returned doc and recurse.
    my $inc = $doc->Root();
    
    $self->_resolve_includes( $inc );

    die "id of including ", $e->Type(), " element (", $e->Attrib('id'),
        " ) doesn't match id of included element (", $inc->Attrib('id'), ")\n"
           if $e->HasAttrib('id') and $inc->HasAttrib('id')
              and $e->Attrib('id') ne $inc->Attrib('id');

    #   4. fixup the attributes.
    # Assumption: including attributes override included attributes.
    foreach my $a ($e->AttribNames)
     {
      next if 'include' eq $a;    # skip the include attribute
      $inc->Attrib( $a => $e->Attrib($a) );
     }

    #   5. take the finished root and replace the element
    $root->ReplaceContent( $e, $inc );
   }

  $self->{doc};
 }



#------------------------------------------------------------
#  Return the outer-most website object
sub   website
 {
  my $self = shift;

  $self->{doc}->Root();
 }


#------------------------------------------------------------
#  Return a reference to the object referenced by the supplied
#  id.
sub   byID
 {
  my $self = shift;
  my $id   = shift;

  defined $id and exists $self->{IDs}->{$id} ? $self->{IDs}->{$id} : undef;
 }


#------------------------------------------------------------
#  Return a list of references to all elements that have IDs and
#  match the given type.
sub   IDedElementsByType
 {
  my $self = shift;
  my $type = shift;

  map { $self->{IDs}->{$_} }
   grep { $type eq $self->{IDs}->{$_}->Type() }
    keys %{$self->{IDs}};
 }



sub   validate
 {
  my $self = shift;
  
  _verify_element( $self->{doc}->Root(), $self->{IDs} );
 }



#---------------------------------------------
#  Test elements for match to WSDL DTD
sub  _verify_element
 {
  my $elem = shift;
  my $type = $elem->Type();
  my $rIDs = shift;
  my $desc = $LegalTags{$type};

  die "Illegal element type \'$type\'.\n" unless defined $desc;

  _validate_attribs( $elem, $rIDs );

  # include case is already resolved by the time we reach this.

  # test content
  my $e;
  foreach $e ($elem->Content( '*' ))
   {
    my $etype = $e->Type();
    die "Unexpected element type \'$etype\' in content for \'$type\'.\n"
                                        unless exists $desc->{content}->{$etype};
    _verify_element( $e, $rIDs );
   }

  foreach $e (grep { $desc->{content}->{$_} } keys %{$desc->{content}})
   {
    die "Missing element type \'$e\' in content for \'$type\'.\n"
                                                unless $elem->Content( $e );
   }

  1;
 }


#  Test attributes for match to WSDL DTD
sub   _validate_attribs
 {
  my $el    = shift;
  my $type  = $el->Type();
  my $rIDs  = shift;
  my $ident = '';

  $ident   = $el->Attrib('id')       if $el->HasAttrib('id');
  $ident ||= $el->Attrib('name')     if $el->HasAttrib('name');
  $ident ||= $el->Attrib('title')    if $el->HasAttrib('title');
  $ident ||= $el->Attrib('ref')      if $el->HasAttrib('ref');
  $ident ||= $el->Attrib('fragment') if $el->HasAttrib('fragment');
  $ident ||=  "unknown"; 

  my $tag = $LegalTags{$type};

  die "Unrecognized tag type \'$type\'.\n" unless defined $tag;

  foreach $a (@{$tag->{required}})
   {
      die "$type($ident): Missing $a attribute.\n" unless $el->HasAttrib($a);
   }

  foreach $a ($el->AttribNames)
   {
    unless(exists $tag->{attr}->{$a})
     {
      die "$type($ident): unrecognized attribute $a.\n";
     }
    elsif($tag->{attr}->{$a})
     {
      die "$type($ident): element referenced by $a attribute ($el->{$a}) not found.\n"
                                            if !defined $rIDs->{$el->{$a}};
      die "$type($ident): element referenced by $a attribute is ($el->{$a}) is not a $tag->{attr}->{$a}.\n"
                                            if $rIDs->{$el->{$a}}->Type() ne $tag->{attr}->{$a};
     }
   }
 }

1;
